import itertools
from collections import OrderedDict
from itertools import product

from pybaum.config import IS_JAX_INSTALLED
from pybaum.config import IS_NUMPY_INSTALLED
from pybaum.config import IS_PANDAS_INSTALLED

if IS_NUMPY_INSTALLED:
    import numpy as np

if IS_PANDAS_INSTALLED:
    import pandas as pd

if IS_JAX_INSTALLED:
    import jax


def _none():
    """Create registry entry for NoneType."""
    entry = {
        type(None): {
            "flatten": lambda tree: ([], None),  # noqa: U100
            "unflatten": lambda aux_data, children: None,  # noqa: U100
            "names": lambda tree: [],  # noqa: U100
        }
    }
    return entry


def _list():
    """Create registry entry for list."""
    entry = {
        list: {
            "flatten": lambda tree: (tree, None),
            "unflatten": lambda aux_data, children: children,  # noqa: U100
            "names": lambda tree: [f"{i}" for i in range(len(tree))],
        },
    }
    return entry


def _dict():
    """Create registry entry for dict."""
    entry = {
        dict: {
            "flatten": lambda tree: (list(tree.values()), list(tree)),
            "unflatten": lambda aux_data, children: dict(zip(aux_data, children)),
            "names": lambda tree: list(map(str, list(tree))),
        },
    }
    return entry


def _tuple():
    """Create registry entry for tuple."""
    entry = {
        tuple: {
            "flatten": lambda tree: (list(tree), None),
            "unflatten": lambda aux_data, children: tuple(children),  # noqa: U100
            "names": lambda tree: [f"{i}" for i in range(len(tree))],
        },
    }
    return entry


def _namedtuple():
    """Create registry entry for namedtuple and NamedTuple."""
    entry = {
        "namedtuple": {
            "flatten": lambda tree: (list(tree), tree),
            "unflatten": _unflatten_namedtuple,
            "names": lambda tree: list(tree._fields),
        },
    }
    return entry


def _unflatten_namedtuple(aux_data, leaves):
    replacements = dict(zip(aux_data._fields, leaves))
    out = aux_data._replace(**replacements)
    return out


def _ordereddict():
    """Create registry entry for OrderedDict."""
    entry = {
        OrderedDict: {
            "flatten": lambda tree: (list(tree.values()), list(tree)),
            "unflatten": lambda aux_data, children: OrderedDict(
                zip(aux_data, children)
            ),
            "names": lambda tree: list(map(str, list(tree))),
        },
    }
    return entry


def _numpy_array():
    """Create registry entry for numpy.ndarray."""

    if IS_NUMPY_INSTALLED:
        entry = {
            np.ndarray: {
                "flatten": lambda arr: (arr.flatten().tolist(), arr.shape),
                "unflatten": lambda aux_data, leaves: np.array(leaves).reshape(
                    aux_data
                ),
                "names": _array_element_names,
            },
        }
    else:
        entry = {}
    return entry


def _array_element_names(arr):
    dim_names = [map(str, range(n)) for n in arr.shape]
    names = list(map("_".join, itertools.product(*dim_names)))
    return names


def _jax_array():
    if IS_JAX_INSTALLED:
        entry = {
            "jax.numpy.ndarray": {
                "flatten": lambda arr: (arr.flatten().tolist(), arr.shape),
                "unflatten": lambda aux_data, leaves: jax.numpy.array(leaves).reshape(
                    aux_data
                ),
                "names": _array_element_names,
            },
        }
    else:
        entry = {}
    return entry


def _pandas_series():
    """Create registry entry for pandas.Series."""
    if IS_PANDAS_INSTALLED:
        entry = {
            pd.Series: {
                "flatten": lambda sr: (
                    sr.tolist(),
                    {"index": sr.index, "name": sr.name},
                ),
                "unflatten": lambda aux_data, leaves: pd.Series(leaves, **aux_data),
                "names": lambda sr: list(sr.index.map(_index_element_to_string)),
            },
        }
    else:
        entry = {}
    return entry


def _pandas_dataframe():
    """Create registry entry for pandas.DataFrame."""
    if IS_PANDAS_INSTALLED:
        entry = {
            pd.DataFrame: {
                "flatten": _flatten_pandas_dataframe,
                "unflatten": _unflatten_pandas_dataframe,
                "names": _get_names_pandas_dataframe,
            }
        }
    else:
        entry = {}
    return entry


def _flatten_pandas_dataframe(df):
    flat = df.to_numpy().flatten().tolist()
    aux_data = {"columns": df.columns, "index": df.index, "shape": df.shape}
    return flat, aux_data


def _unflatten_pandas_dataframe(aux_data, leaves):
    out = pd.DataFrame(
        data=np.array(leaves).reshape(aux_data["shape"]),
        columns=aux_data["columns"],
        index=aux_data["index"],
    )
    return out


def _get_names_pandas_dataframe(df):
    index_strings = list(df.index.map(_index_element_to_string))
    out = ["_".join([loc, col]) for loc, col in product(index_strings, df.columns)]
    return out


def _index_element_to_string(element):
    if isinstance(element, (tuple, list)):
        as_strings = [str(entry) for entry in element]
        res_string = "_".join(as_strings)
    else:
        res_string = str(element)

    return res_string


FUNC_DICT = {
    "list": _list,
    "tuple": _tuple,
    "dict": _dict,
    "numpy.ndarray": _numpy_array,
    "jax.numpy.ndarray": _jax_array,
    "pandas.Series": _pandas_series,
    "pandas.DataFrame": _pandas_dataframe,
    "None": _none,
    "namedtuple": _namedtuple,
    "OrderedDict": _ordereddict,
}



from pybaum.registry_entries import FUNC_DICT


def get_registry(types=None, include_defaults=True):
    """Create a pytree registry.

    Args:
        types (list): A list strings with the names of types that should be included in
            the registry, i.e. considered containers and not leaves by the functions
            that work with pytrees. Currently we support:
            - "tuple"
            - "dict"
            - "list"
            - :class:`collections.namedtuple` or :class:`typing.NamedTuple`
            - :obj:`None`
            - :class:`collections.OrderedDict`
            - "numpy.ndarray"
            - "jax.numpy.ndarray"
            - "pandas.Series"
            - "pandas.DataFrame"
        include_defaults (bool): Whether the default pytree containers "tuple", "dict"
            "list", "None", "namedtuple" and "OrderedDict" should be included even if
            not specified in `types`.

    Returns:
        dict: A pytree registry.

    """
    types = [] if types is None else types

    if include_defaults:
        default_types = {"list", "tuple", "dict", "None", "namedtuple", "OrderedDict"}
        types = list(set(types) | default_types)

    registry = {}
    for typ in types:
        new_entry = FUNC_DICT[typ]()
        registry = {**registry, **new_entry}

    return registry



"""Implement functionality similar to jax.tree_util in pure Python.

The functions are not completely identical to jax. The most notable differences are:

- Instead of a global registry of pytree nodes, most functions have a registry argument.
- The treedef containing information to unflatten pytrees is implemented differently.

"""
from pybaum.equality import EQUALITY_CHECKERS
from pybaum.registry import get_registry
from pybaum.typecheck import get_type


def tree_flatten(tree, is_leaf=None, registry=None):
    """Flatten a pytree and create a treedef.

    Args:
        tree: a pytree to flatten.
        is_leaf (callable or None): An optionally specified function that will be called
            at each flattening step. It should return a boolean, which indicates whether
            the flattening should traverse the current object, or if it should be
            stopped immediately, with the whole subtree being treated as a leaf.
        registry (dict or None): A pytree container registry that determines
            which types are considered container objects that should be flattened.
            ``is_leaf`` can override this in the sense that types that are in the
            registry are still considered a leaf but it cannot declare something a
            container that is not in the registry. None means that the default registry
            is used, i.e. that dicts, tuples and lists are considered containers.
            "extended" means that in addition numpy arrays and params DataFrames are
            considered containers. Passing a dictionary where the keys are types and the
            values are dicts with the entries "flatten", "unflatten" and "names" allows
            to completely override the default registries.

    Returns:
        A pair where the first element is a list of leaf values and the second
        element is a treedef representing the structure of the flattened tree.

    """
    registry = _process_pytree_registry(registry)
    is_leaf = _process_is_leaf(is_leaf)

    flat = _tree_flatten(tree, is_leaf=is_leaf, registry=registry)
    # unflatten the flat tree to make a copy
    treedef = tree_unflatten(tree, flat, is_leaf=is_leaf, registry=registry)
    return flat, treedef


def tree_just_flatten(tree, is_leaf=None, registry=None):
    """Flatten a pytree without creating a treedef.

    Args:
        tree: a pytree to flatten.
        is_leaf (callable or None): An optionally specified function that will be called
            at each flattening step. It should return a boolean, which indicates whether
            the flattening should traverse the current object, or if it should be
            stopped immediately, with the whole subtree being treated as a leaf.
        registry (dict or None): A pytree container registry that determines
            which types are considered container objects that should be flattened.
            ``is_leaf`` can override this in the sense that types that are in the
            registry are still considered a leaf but it cannot declare something a
            container that is not in the registry. None means that the default registry
            is used, i.e. that dicts, tuples and lists are considered containers.
            "extended" means that in addition numpy arrays and params DataFrames are
            considered containers. Passing a dictionary where the keys are types and the
            values are dicts with the entries "flatten", "unflatten" and "names" allows
            to completely override the default registries.

    Returns:
        A pair where the first element is a list of leaf values and the second
        element is a treedef representing the structure of the flattened tree.

    """
    registry = _process_pytree_registry(registry)
    is_leaf = _process_is_leaf(is_leaf)

    flat = _tree_flatten(tree, is_leaf=is_leaf, registry=registry)
    return flat


def _tree_flatten(tree, is_leaf, registry):
    out = []
    tree_type = get_type(tree)

    if tree_type not in registry or is_leaf(tree):
        out.append(tree)
    else:
        subtrees, _ = registry[tree_type]["flatten"](tree)
        for subtree in subtrees:
            if get_type(subtree) in registry:
                out += _tree_flatten(subtree, is_leaf, registry)
            else:
                out.append(subtree)
    return out


def tree_yield(tree, is_leaf=None, registry=None):
    """Yield leafs from a pytree and create the tree definition.

    Args:
        tree: a pytree.
        is_leaf (callable or None): An optionally specified function that will be called
            at each yield step. It should return a boolean, which indicates whether
            the generator should traverse the current object, or if it should be
            stopped immediately, with the whole subtree being treated as a leaf.
        registry (dict or None): A pytree container registry that determines
            which types are considered container objects that should be yielded.
            ``is_leaf`` can override this in the sense that types that are in the
            registry are still considered a leaf but it cannot declare something a
            container that is not in the registry. None means that the default registry
            is used, i.e. that dicts, tuples and lists are considered containers.
            "extended" means that in addition numpy arrays and params DataFrames are
            considered containers. Passing a dictionary where the keys are types and the
            values are dicts with the entries "flatten", "unflatten" and "names" allows
            to completely override the default registries.

    Returns:
        A pair where the first element is a generator of leaf values and the second
        element is a treedef representing the structure of the flattened tree.

    """
    registry = _process_pytree_registry(registry)
    is_leaf = _process_is_leaf(is_leaf)

    flat = _tree_yield(tree, is_leaf=is_leaf, registry=registry)
    return flat, tree


def tree_just_yield(tree, is_leaf=None, registry=None):
    """Yield leafs from a pytree without creating a treedef.

    Args:
        tree: a pytree.
        is_leaf (callable or None): An optionally specified function that will be called
            at each yield step. It should return a boolean, which indicates whether
            the generator should traverse the current object, or if it should be
            stopped immediately, with the whole subtree being treated as a leaf.
        registry (dict or None): A pytree container registry that determines
            which types are considered container objects that should be yielded.
            ``is_leaf`` can override this in the sense that types that are in the
            registry are still considered a leaf but it cannot declare something a
            container that is not in the registry. None means that the default registry
            is used, i.e. that dicts, tuples and lists are considered containers.
            "extended" means that in addition numpy arrays and params DataFrames are
            considered containers. Passing a dictionary where the keys are types and the
            values are dicts with the entries "flatten", "unflatten" and "names" allows
            to completely override the default registries.

    Returns:
        A generator of leaf values.

    """
    registry = _process_pytree_registry(registry)
    is_leaf = _process_is_leaf(is_leaf)

    flat = _tree_yield(tree, is_leaf=is_leaf, registry=registry)
    return flat


def _tree_yield(tree, is_leaf, registry):
    out = []
    tree_type = get_type(tree)

    if tree_type not in registry or is_leaf(tree):
        yield tree
    else:
        subtrees, _ = registry[tree_type]["flatten"](tree)
        for subtree in subtrees:
            if get_type(subtree) in registry:
                yield from _tree_yield(subtree, is_leaf, registry)
            else:
                yield subtree
    return out


def tree_unflatten(treedef, leaves, is_leaf=None, registry=None):
    """Reconstruct a pytree from the treedef and a list of leaves.

    The inverse of :func:`tree_flatten`.

    Args:
        treedef: the treedef to with information needed for reconstruction.
        leaves (list): the list of leaves to use for reconstruction. The list must match
            the leaves of the treedef.
        is_leaf (callable or None): An optionally specified function that will be called
            at each flattening step. It should return a boolean, which indicates whether
            the flattening should traverse the current object, or if it should be
            stopped immediately, with the whole subtree being treated as a leaf.
        registry (dict or None): A pytree container registry that determines
            which types are considered container objects that should be flattened.
            `is_leaf` can override this in the sense that types that are in the
            registry are still considered a leaf but it cannot declare something a
            container that is not in the registry. None means that the default registry
            is used, i.e. that dicts, tuples and lists are considered containers.
            "extended" means that in addition numpy arrays and params DataFrames are
            considered containers. Passing a dictionary where the keys are types and the
            values are dicts with the entries "flatten", "unflatten" and "names" allows
            to completely override the default registries.

    Returns:
        The reconstructed pytree, containing the ``leaves`` placed in the structure
        described by ``treedef``.

    """
    registry = _process_pytree_registry(registry)
    is_leaf = _process_is_leaf(is_leaf)
    return _tree_unflatten(treedef, leaves, is_leaf=is_leaf, registry=registry)


def _tree_unflatten(treedef, leaves, is_leaf, registry):
    leaves = iter(leaves)
    tree_type = get_type(treedef)

    if tree_type not in registry or is_leaf(treedef):
        return next(leaves)
    else:
        items, info = registry[tree_type]["flatten"](treedef)
        unflattened_items = []
        for item in items:
            if get_type(item) in registry:
                unflattened_items.append(
                    _tree_unflatten(item, leaves, is_leaf=is_leaf, registry=registry)
                )
            else:
                unflattened_items.append(next(leaves))
        return registry[tree_type]["unflatten"](info, unflattened_items)


def tree_map(func, tree, is_leaf=None, registry=None):
    """Apply func to all leaves in tree.

    Args:
        func (callable): Function applied to each leaf in the tree.
        tree: A pytree.
        is_leaf (callable or None): An optionally specified function that will be called
            at each flattening step. It should return a boolean, which indicates whether
            the flattening should traverse the current object, or if it should be
            stopped immediately, with the whole subtree being treated as a leaf.
        registry (dict or None): A pytree container registry that determines
            which types are considered container objects that should be flattened.
            `is_leaf` can override this in the sense that types that are in the
            registry are still considered a leaf but it cannot declare something a
            container that is not in the registry. None means that the default registry
            is used, i.e. that dicts, tuples and lists are considered containers.
            "extended" means that in addition numpy arrays and params DataFrames are
            considered containers. Passing a dictionary where the keys are types and the
            values are dicts with the entries "flatten", "unflatten" and "names" allows
            to completely override the default registries.
    Returns:
        modified copy of tree.

    """
    flat, treedef = tree_flatten(tree, is_leaf=is_leaf, registry=registry)
    modified = [func(i) for i in flat]
    new_tree = tree_unflatten(treedef, modified, is_leaf=is_leaf, registry=registry)
    return new_tree


def tree_multimap(func, *trees, is_leaf=None, registry=None):
    """Apply func to leaves of multiple pytrees.

    Args:
        func (callable): Function applied to each leaf corresponding leaves of
            multiple py trees.
        trees: An arbitrary number of pytrees. All trees need to have the same
            structure.
        is_leaf (callable or None): An optionally specified function that will be called
            at each flattening step. It should return a boolean, which indicates whether
            the flattening should traverse the current object, or if it should be
            stopped immediately, with the whole subtree being treated as a leaf.
        registry (dict or None): A pytree container registry that determines
            which types are considered container objects that should be flattened.
            `is_leaf` can override this in the sense that types that are in the
            registry are still considered a leaf but it cannot declare something a
            container that is not in the registry. None means that the default registry
            is used, i.e. that dicts, tuples and lists are considered containers.
            "extended" means that in addition numpy arrays and params DataFrames are
            considered containers. Passing a dictionary where the keys are types and the
            values are dicts with the entries "flatten", "unflatten" and "names" allows
            to completely override the default registries.
    Returns:
        tree with the same structure as the elements in trees.

    """
    flat_trees, treedefs = [], []
    for tree in trees:
        flat, treedef = tree_flatten(tree, is_leaf=is_leaf, registry=registry)
        flat_trees.append(flat)
        treedefs.append(treedef)

    for treedef in treedefs:
        if treedef != treedefs[0]:
            raise ValueError("All trees must have the same structure.")

    modified = [func(*item) for item in zip(*flat_trees)]

    new_trees = tree_unflatten(
        treedefs[0], modified, is_leaf=is_leaf, registry=registry
    )
    return new_trees


def leaf_names(tree, is_leaf=None, registry=None, separator="_"):
    """Construct names for leaves in a pytree.

    Args:
        tree: a pytree to flatten.
        is_leaf (callable or None): An optionally specified function that will be called
            at each flattening step. It should return a boolean, which indicates whether
            the flattening should traverse the current object, or if it should be
            stopped immediately, with the whole subtree being treated as a leaf.
        registry (dict or None): A pytree container registry that determines
            which types are considered container objects that should be flattened.
            `is_leaf` can override this in the sense that types that are in the
            registry are still considered a leaf but it cannot declare something a
            container that is not in the registry. None means that the default registry
            is used, i.e. that dicts, tuples and lists are considered containers.
            "extended" means that in addition numpy arrays and params DataFrames are
            considered containers. Passing a dictionary where the keys are types and the
            values are dicts with the entries "flatten", "unflatten" and "names" allows
            to completely override the default registries.
        separator (str): String that separates the building blocks of the leaf name.
    Returns:
        list: List of strings with names for pytree leaves.

    """
    registry = _process_pytree_registry(registry)
    is_leaf = _process_is_leaf(is_leaf)
    leaf_names = _leaf_names(
        tree, is_leaf=is_leaf, registry=registry, separator=separator
    )
    return leaf_names


def _leaf_names(tree, is_leaf, registry, separator, prefix=None):
    out = []
    tree_type = get_type(tree)

    if tree_type not in registry or is_leaf(tree):
        out.append(prefix)
    else:
        subtrees, info = registry[tree_type]["flatten"](tree)
        names = registry[tree_type]["names"](tree)
        for name, subtree in zip(names, subtrees):
            if get_type(subtree) in registry:
                out += _leaf_names(
                    subtree,
                    is_leaf=is_leaf,
                    registry=registry,
                    separator=separator,
                    prefix=_add_prefix(prefix, name, separator),
                )
            else:
                out.append(_add_prefix(prefix, name, separator))
    return out


def _add_prefix(prefix, string, separator):
    if prefix not in (None, ""):
        out = separator.join([prefix, string])
    else:
        out = string
    return out


def _process_pytree_registry(registry):
    registry = registry if registry is not None else get_registry()
    return registry


def _process_is_leaf(is_leaf):
    if is_leaf is None:
        return lambda tree: False  # noqa: U100
    else:
        return is_leaf


def tree_equal(tree, other, is_leaf=None, registry=None, equality_checkers=None):
    """Determine if two pytrees are equal.

    Two pytrees are considered equal if their leaves are equal and the names of their
    leaves are equal. While this definition of equality might not always make sense
    it makes sense in most cases and can be implemented relatively easily.

    Args:
        tree: A pytree.
        other: Another pytree.
        is_leaf (callable or None): An optionally specified function that will be called
            at each flattening step. It should return a boolean, which indicates whether
            the flattening should traverse the current object, or if it should be
            stopped immediately, with the whole subtree being treated as a leaf.
        registry (dict or None): A pytree container registry that determines
            which types are considered container objects that should be flattened.
            `is_leaf` can override this in the sense that types that are in the
            registry are still considered a leaf but it cannot declare something a
            container that is not in the registry. None means that the default registry
            is used, i.e. that dicts, tuples and lists are considered containers.
            "extended" means that in addition numpy arrays and params DataFrames are
            considered containers. Passing a dictionary where the keys are types and the
            values are dicts with the entries "flatten", "unflatten" and "names" allows
            to completely override the default registries.
        equality_checkers (dict, None): A dictionary where keys are types and values are
            functions which assess equality for the type of object.

    Returns:
        bool

    """
    equality_checkers = (
        EQUALITY_CHECKERS
        if equality_checkers is None
        else {**EQUALITY_CHECKERS, **equality_checkers}
    )

    first_flat = tree_just_flatten(tree, is_leaf=is_leaf, registry=registry)
    second_flat = tree_just_flatten(other, is_leaf=is_leaf, registry=registry)

    first_names = leaf_names(tree, is_leaf=is_leaf, registry=registry)
    second_names = leaf_names(tree, is_leaf=is_leaf, registry=registry)

    equal = first_names == second_names

    if equal:
        for first, second in zip(first_flat, second_flat):
            check_func = equality_checkers.get(get_type(first), lambda a, b: a == b)
            equal = equal and check_func(first, second)
            if not equal:
                break

    return equal


def tree_update(tree, other, is_leaf=None, registry=None):
    """Update leaves in a pytree with leaves from another pytree.

    The second pytree must be compatible with the first one but can be smaller. For
    example, lists can be shorter, dictionaries can contain subsets of entries, etc.

    Args:
        tree: A pytree.
        other: Another pytree.
        is_leaf (callable or None): An optionally specified function that will be called
            at each flattening step. It should return a boolean, which indicates whether
            the flattening should traverse the current object, or if it should be
            stopped immediately, with the whole subtree being treated as a leaf.
        registry (dict or None): A pytree container registry that determines
            which types are considered container objects that should be flattened.
            `is_leaf` can override this in the sense that types that are in the
            registry are still considered a leaf but it cannot declare something a
            container that is not in the registry. None means that the default registry
            is used, i.e. that dicts, tuples and lists are considered containers.
            "extended" means that in addition numpy arrays and params DataFrames are
            considered containers. Passing a dictionary where the keys are types and the
            values are dicts with the entries "flatten", "unflatten" and "names" allows
            to completely override the default registries.
    Returns:
        Updated pytree.

    """
    first_flat, first_treedef = tree_flatten(tree, is_leaf=is_leaf, registry=registry)
    first_names = leaf_names(tree, is_leaf=is_leaf, registry=registry)
    first_dict = dict(zip(first_names, first_flat))

    other_flat, _ = tree_flatten(other, is_leaf=is_leaf, registry=registry)
    other_names = leaf_names(other, is_leaf=is_leaf, registry=registry)
    other_dict = dict(zip(other_names, other_flat))

    combined = list({**first_dict, **other_dict}.values())

    out = tree_unflatten(first_treedef, combined, is_leaf=is_leaf, registry=registry)
    return out



try:
    import numpy as np  # noqa: F401
except ImportError:
    IS_NUMPY_INSTALLED = False
else:
    IS_NUMPY_INSTALLED = True


try:
    import pandas as pd  # noqa: F401
except ImportError:
    IS_PANDAS_INSTALLED = False
else:
    IS_PANDAS_INSTALLED = True


try:
    import jax  # noqa: F401
    import jaxlib  # noqa: F401
except ImportError:
    IS_JAX_INSTALLED = False
else:
    IS_JAX_INSTALLED = True



from pybaum.config import IS_JAX_INSTALLED
from pybaum.config import IS_NUMPY_INSTALLED

if IS_JAX_INSTALLED:
    import jax.numpy as jnp

if IS_NUMPY_INSTALLED:
    import numpy as np


def get_type(obj):
    """Get type of candidate objects in a pytree.

    This function allows us to reliably identify namedtuples, NamedTuples and jax arrays
    for which standard ``type`` function does not work.

    Args:
        obj: The object to be checked

    Returns:
        type or str: The type of the object or a string with the type name.

    """
    if _is_namedtuple(obj):
        out = "namedtuple"
    elif _is_jax_array(obj):
        out = "jax.numpy.ndarray"
    else:
        out = type(obj)
    return out


def _is_namedtuple(obj):
    """Check if an object is a namedtuple.

    As in JAX we treat collections.namedtuple and typing.NamedTuple both as
    namedtuple but the exact type is preserved in the unflatten function.

    namedtuples are discovered by being instances of tuple and having a
    ``_fields`` attribute as suggested by Raymond Hettinger
    `here <https://bugs.python.org/issue7796>`_.

    Moreover we check for the presence of a ``_replace`` method because we need when
    unflattening pytrees.

    This can produce false positives but in most cases would still result in desired
    behavior.

    Args:
        obj: The object to be checked

    Returns:
        bool

    """
    out = (
        isinstance(obj, tuple) and hasattr(obj, "_fields") and hasattr(obj, "_replace")
    )
    return out


def _is_jax_array(obj):
    """Check if an object is a jax array.

    The exact type of jax arrays has changed over time and is an implementation detail.

    Instead we rely on isinstance checks which will likely be more stable in the future.
    However, the behavior of isinstance for jax arrays has also changed over time. For
    jax versions before 0.2.21, standard numpy arrays were instances of jax arrays,
    now they are not.

    Resources:
    ----------

    - https://github.com/google/jax/issues/2115
    - https://github.com/google/jax/issues/2014
    - https://github.com/google/jax/blob/main/CHANGELOG.md#jax-0221-sept-23-2021
    - https://github.com/google/jax/blob/main/CHANGELOG.md#jax-0318-sep-26-2022

    Args:
        obj: The object to be checked

    Returns:
        bool

    """
    if not IS_JAX_INSTALLED:
        out = False
    elif IS_NUMPY_INSTALLED:
        out = isinstance(obj, jnp.ndarray) and not isinstance(obj, np.ndarray)
    else:
        out = isinstance(obj, jnp.ndarray)
    return out



from pybaum.registry import get_registry
from pybaum.tree_util import leaf_names
from pybaum.tree_util import tree_equal
from pybaum.tree_util import tree_flatten
from pybaum.tree_util import tree_just_flatten
from pybaum.tree_util import tree_just_yield
from pybaum.tree_util import tree_map
from pybaum.tree_util import tree_multimap
from pybaum.tree_util import tree_unflatten
from pybaum.tree_util import tree_update
from pybaum.tree_util import tree_yield


__all__ = [
    "tree_flatten",
    "tree_just_flatten",
    "tree_just_yield",
    "tree_unflatten",
    "tree_map",
    "tree_multimap",
    "leaf_names",
    "tree_equal",
    "tree_update",
    "tree_yield",
    "get_registry",
]



"""Functions to check equality of pytree leaves."""
from pybaum.config import IS_JAX_INSTALLED
from pybaum.config import IS_NUMPY_INSTALLED
from pybaum.config import IS_PANDAS_INSTALLED


if IS_NUMPY_INSTALLED:
    import numpy as np


if IS_PANDAS_INSTALLED:
    import pandas as pd

EQUALITY_CHECKERS = {}


if IS_NUMPY_INSTALLED:
    EQUALITY_CHECKERS[np.ndarray] = lambda a, b: bool((a == b).all())


if IS_PANDAS_INSTALLED:
    EQUALITY_CHECKERS[pd.Series] = lambda a, b: a.equals(b)
    EQUALITY_CHECKERS[pd.DataFrame] = lambda a, b: a.equals(b)


if IS_JAX_INSTALLED:
    EQUALITY_CHECKERS["jax.numpy.ndarray"] = lambda a, b: bool((a == b).all())
